#
#     Copyright (C) 2007 Martyn Winn
#
#     This code is distributed under the terms and conditions of the
#     CCP4 Program Suite Licence Agreement as a CCP4 Library.
#     A copy of the CCP4 licence can be obtained by writing to the
#     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
#
#CCP4i_cvs_Id $Id: pisa.tcl,v 1.7 2008/03/11 12:26:32 pjx Exp $
# ======================================================================
# pisa.tcl --
#
# Run the command line PISA to get information on interfaces and multimers
#
# CCP4Interface 
#
# =======================================================================

#==========================================================================
#==_setup==initialisations, defining local menus===========================
#==========================================================================

proc pisa_setup { typedefname arrayname } {

    upvar #0 $typedefname typedef
    
    set typedef(_pisa_input) { menu { "Generate new PISA analysis" "Use previous PISA analysis" } { ANALYSE REVIEW } }
    
    set typedef(_pisa_option) { menu { "list interfaces, monomers and assemblies" \
                                       "view interface, monomer, assembly or dissociate" \
                                       "download interface, monomer, assembly or dissociate" \
                                       "get details of an interface, monomer or assembly" \
                                       "generate XML output" \
                                        } { LIST VIEW DOWNLOAD DETAIL XML } }

    set typedef(_pisa_spec) { menu { "interface" \
                                       "monomer" \
                                       "assembly" \
                                       "dissociate" \
                                        } { INTERFACE MONOMER ASSEMBLY DISSOCIATE } }

    return 1
    
}

#==========================================================================
#==_task_window==core procedure. all commands to define graphical interface
#==========================================================================

proc pisa_task_window { arrayname } {

    upvar #0 $arrayname array
    
    if { [CreateTaskWindow $arrayname \
	      "Analyse protein interfaces" "Pisa" [list "Generate new PISA analysis" "Use previous PISA analysis" ] \
	      {} \
	      -action_buttons [ list quit interactive ] ] == 0 } return

#==========================================================================
#==PROTOCOL==TITLE AND KEY DECISIONS=======================================
#==========================================================================
    
    OpenFolder protocol
    
    CreateTitleLine line TITLE
    
    CreateLine line \
        message "Generate new interface information, or re-visit a previous job" \
        widget INPUT 

#=========================================================================
#==FOLDER 1==Run PISA and generate output directory of results  ==========
#=========================================================================
    
    OpenFolder 1 INPUT open ANALYSE hide

    CreateInputFileLine line \
        "Enter name of input pdb file" \
        "PDB in" \
        XYZIN DIR_XYZIN 

#===========================================================================
#==FOLDER 2==Use results directory from previous run =======================
#===========================================================================

    OpenFolder 2 INPUT open REVIEW hide

    CreateLine line label "Review previously generated results, from job " \
       widget NAME \
       label " of the current project"
    
    CreateLine line \
        message "Choose which mode of PISA to run" \
        widget OPTION

    CreateLine line \
        label "for " \
        widget SPEC \
        label "number " \
        widget SERIAL_NO \
        toggle_display OPTION open [list "VIEW" "DOWNLOAD" "DETAIL"]

    CreateOutputFileLine line \
        "Output coordinate file from download option" \
        "PDB out" \
        XYZOUT DIR_XYZOUT \
        -toggle_display OPTION open { DOWNLOAD }

    CreateOutputFileLine line \
        "Output XML file" \
        "XML out" \
        XMLOUT DIR_XMLOUT \
        -toggle_display OPTION open { XML }

}

#==========================================================================
#==_run==executed when user hits .....               ======================
#==========================================================================

proc pisa_run { arrayname } {

    upvar #0 $arrayname array
    global env

# Set PISA_WORK_ROOT to the ccp4i project directory
# This is where the output of the analysis step will be stored.
    set env(PISA_WORK_ROOT) [ GetDefaultDirPath ]

    set session_name "myname" 
    if { [GetValue $arrayname INPUT] == "ANALYSE" } {
	set session_name [expr [DbGetNJOBS] +1]
    } elseif { $array(NAME) != "" } {
      set session_name $array(NAME)
    } else {
      PleaseWait
      WarningMessage "No session name entered"
      return 0
    }      

# Set log file, and write header information
    set array(LOGFILE) [GetTmpFileName]
    set log [GetTmpFileName]
    set logheader [WriteIdentifier {} "LOG pisa" \
	SCRATCH [GetEnvPath CCP4_SCR] \
	HOSTNAME [GetHostname] \
	PID [GetPid] ]
    WriteFile $array(LOGFILE) $logheader
    WriteFile $array(LOGFILE) "\n \#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#"
    WriteFile $array(LOGFILE)   " \#\#\#\#\#\#\#   PISA INTERACTIVE TASK    \#\#\#\#\#\#\#\#\#\#\#\#\#\#\#"
    WriteFile $array(LOGFILE)   " \#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\n"

   if { [GetValue $arrayname INPUT] == "ANALYSE" } {
     set cmd "pisa $session_name "
     SetDefaultTitle $arrayname "new PISA session"
     append cmd " -analyse " 
     append cmd "[GetFullFileName0 $arrayname XYZIN] " 
  
     WriteFile $log "Running command \"$cmd\""
     if { ![Execute $cmd {} status report -log $log ] } {
       PleaseWait
       WarningMessage "Error running pisa program (analyse option) - see job log file"
       if { [file exists $log] } { TranscribeFile $log $array(LOGFILE) }
       return 0
     }
   }

    if { ([GetValue $arrayname INPUT] == "ANALYSE") \
	 || ( ([GetValue $arrayname INPUT] == "REVIEW") \
	    && ([GetValue $arrayname OPTION] == "LIST") ) } {
# use -list command to summarise results

     set cmd "pisa $session_name "
     SetDefaultTitle $arrayname "review PISA session"
     append cmd "-list interfaces " 
  
     WriteFile $log "Running command \"$cmd\"" 
     if { ![Execute $cmd {} status report -log $log ] } {
       PleaseWait
       WarningMessage "Error running pisa program (list interfaces option) - see job log file"
       if { [file exists $log] } { TranscribeFile $log $array(LOGFILE) }
       return 0
     }

     set cmd "pisa $session_name "
     SetDefaultTitle $arrayname "review PISA session"
     append cmd "-list monomers " 
  
     WriteFile $log "Running command \"$cmd\"" 
     if { ![Execute $cmd {} status report -log $log ] } {
       PleaseWait
       WarningMessage "Error running pisa program (list monomers option) - see job log file"
       if { [file exists $log] } { TranscribeFile $log $array(LOGFILE) }
       return 0
     }

     set cmd "pisa $session_name "
     SetDefaultTitle $arrayname "review PISA session"
     append cmd "-list assemblies " 
  
     WriteFile $log "Running command \"$cmd\"" 
     if { ![Execute $cmd {} status report -log $log ] } {
       PleaseWait
       WarningMessage "Error running pisa program (list assemblies option) - see job log file"
       if { [file exists $log] } { TranscribeFile $log $array(LOGFILE) }
       return 0
     }
     if { [file exists $log] } { TranscribeFile $log $array(LOGFILE) }
   }

   if { ([GetValue $arrayname INPUT] == "REVIEW") &&   \
        ([GetValue $arrayname OPTION] == "VIEW") } {
# use -view command to view an interface, monomer or assembly
# at the moment this is Rasmol

     set cmd "pisa $session_name "
     SetDefaultTitle $arrayname "review PISA session"
     append cmd "-view $array(SPEC) $array(SERIAL_NO) " 
  
     WriteFile $log "Running command \"$cmd\"" 
     if { ![Execute $cmd {} status report -log $log ] } {
       PleaseWait
       WarningMessage "Error running pisa program (view option) - see job log file"
       if { [file exists $log] } { TranscribeFile $log $array(LOGFILE) }
       return 0
     }
     if { [file exists $log] } { TranscribeFile $log $array(LOGFILE) }
   }

   if { ([GetValue $arrayname INPUT] == "REVIEW") &&   \
        ([GetValue $arrayname OPTION] == "DOWNLOAD") } {
# use -download command to download an interface, monomer or assembly

     set xyz_file [GetTmpFileName]
     set cmd "pisa $session_name "
     SetDefaultTitle $arrayname "download PISA session"
     append cmd "-download $array(SPEC) $array(SERIAL_NO)" 
  
     WriteFile $array(LOGFILE) "Running command \"$cmd\"" 
     if { ![Execute $cmd {} status report -log $xyz_file ] } {
       PleaseWait
       WarningMessage "Error running pisa program (download option) - see job log file"
       if { [file exists $xyz_file] } { TranscribeFile $xyz_file $array(LOGFILE) }
       WriteFile $array(LOGFILE) "PISA failed to produced a download file." 
       return 0
     }
# in this case, the log file is actually the requested download file
     set array(OUTPUT_FILES) "XYZOUT"
     if { [file exists $xyz_file] } { MoveFile $xyz_file [GetFullFileName $array(XYZOUT) $array(DIR_XYZOUT)] }
     DeleteFile $xyz_file
     WriteFile $array(LOGFILE) "PISA successfully produced a download file." 
   }

   if { ([GetValue $arrayname INPUT] == "REVIEW") &&   \
        ([GetValue $arrayname OPTION] == "DETAIL") } {
# use -detail command to get details of an interface, monomer or assembly

     set cmd "pisa $session_name "
     SetDefaultTitle $arrayname "review PISA session"
     append cmd "-detail $array(SPEC) $array(SERIAL_NO) " 
  
     WriteFile $log "Running command \"$cmd\"" 
     if { ![Execute $cmd {} status report -log $log ] } {
       PleaseWait
       WarningMessage "Error running pisa program (detail option) - see job log file"
       if { [file exists $log] } { TranscribeFile $log $array(LOGFILE) }
       return 0
     }
     if { [file exists $log] } { TranscribeFile $log $array(LOGFILE) }
   }

   if { ([GetValue $arrayname INPUT] == "REVIEW") &&   \
        ([GetValue $arrayname OPTION] == "XML") } {
# use -xml command to get xml output file

     set xml_file [GetTmpFileName]
     set cmd "pisa $session_name "
     SetDefaultTitle $arrayname "XML PISA session"
     append cmd "-xml " 
  
     WriteFile $array(LOGFILE) "Running command \"$cmd\"" 
     if { ![Execute $cmd {} status report -log $xml_file ] } {
       PleaseWait
       WarningMessage "Error running pisa program (xml option) - see job log file"
       if { [file exists $xml_file] } { TranscribeFile $xml_file $array(LOGFILE) }
       WriteFile $array(LOGFILE) "PISA failed to produced an XML file." 
       return 0
     }
# in this case, the log file is actually the requested XML file
     set array(OUTPUT_FILES) "XMLOUT"
     if { [file exists $xml_file] } { MoveFile $xml_file [GetFullFileName $array(XMLOUT) $array(DIR_XMLOUT)] }
     DeleteFile $xml_file
     WriteFile $array(LOGFILE) "PISA successfully produced an XML file." 
   }

   return 1

}

